# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import PGUtils

NEW = 1
MOVE = 2
REMOVE = 3
CHANGE = 4

GROW = 1
FADE = 2

DEFAULT_DURATION = 0.5

class Animation(object):
    def __init__(self, position=None, segments=None):
        self.position = position
        self.segments = sorted(segments, key=lambda s: s.start_time) if segments else []
        
    def start_time(self):
        return self.segments[0].start_time if self.segments else 0.0
        
    def end_time(self):
        return self.segments[-1].end_time if self.segments else 0.0
        
    def reversed(self, duration):
        pass
        
class AnimationSegment(object):
    def __init__(self, **kwargs):
        for param in ('action', 'animation', 'from_position', 'to_position', 'from_value', 'to_value', 'start_time', 'end_time'):
            setattr(self, param, kwargs.get(param))
            
    def apply_animation(self, pos, value, scale, t):
        """ Returns (current_pos, (from_val, to_val, fraction), scale) """
        fraction = 0.0
        if t>=self.end_time:
            fraction = 1.0
        elif t>self.start_time:
            fraction = (t-self.start_time) / (self.end_time-self.start_time)
            
        next_pos = self.to_position if self.to_position else pos
        next_value = self.to_value if (self.to_value is not None) else value
        next_scale = scale
        if self.action==NEW:
            next_scale = 1.0
        elif self.action==REMOVE:
            next_scale = 0.0
            
        return (blended_array(pos, next_pos, fraction), (value, next_value, fraction), scale + fraction*(next_scale-scale))
        
    def reversed(self, duration):
        rseg = AnimationSegment()
        rseg.from_position = self.to_position
        rseg.to_position = self.from_position
        rseg.from_value = self.to_value
        rseg.to_value = self.from_value
        rseg.start_time = duration - self.end_time
        rseg.end_time = duration - self.start_time
        rseg.animation = self.animation
        rseg.action = self.action
        if rseg.action==NEW: rseg.action = REMOVE
        elif rseg.action==REMOVE: rseg.action = NEW
        return rseg
        
def compute_animation_state(segments, pos, initial_value, scale, t):
    for aseg in segments:
        if t>=aseg.end_time:
            # apply fully
            pos, (_, initial_value, _), scale = aseg.apply_animation(pos, initial_value, scale, t)
        elif t>aseg.start_time:
            # return the in-progress animation state
            return aseg.apply_animation(pos, initial_value, scale, t)
    # return (completed) animation state
    return (pos, (initial_value, initial_value, 0), scale)

class AnimationGroup(object):
    def __init__(self, animations=None):
        self.animations = sorted(animations, key=lambda a: (a.start_time(), a.end_time())) if animations else []
        self.animation_dict = dict((a.position, a) for a in self.animations if a.position)

    def duration(self):
        return max(a.end_time() for a in self.animations) if self.animations else 0
        
    def animation_state_at_time(self, pos, initial_value, t):
        """ Returns (current_pos, (from_val, to_val, fraction), scale) """
        if pos not in self.animation_dict: return (pos, (initial_value, initial_value, 0), 1.0)
        return compute_animation_state(self.animation_dict[pos].segments, pos, initial_value, 1.0, t)
        
    def new_states_at_time(self, t):
        newanims = (a for a in self.animations if a.position is None and t>a.start_time())
        for a in newanims:
            # get starting position and value from first animation segment
            yield compute_animation_state(a.segments, a.segments[0].from_position, a.segments[0].from_value, 0.0, t)

        
def blended_array(a1, a2, fraction):
    # type of returned sequence is same as a1
    return type(a1)(x*(1-fraction) + y*fraction for x,y in zip(a1,a2))
    
def move_animation(frompos, topos, start_time=0.0, duration=DEFAULT_DURATION):
    segment = AnimationSegment(action=MOVE, to_position=topos, start_time=start_time, end_time=start_time+duration)
    return Animation(position=frompos, segments=[segment])
    
def remove_animation(pos, start_time=0.0, duration=DEFAULT_DURATION):
    segment = AnimationSegment(action=REMOVE, start_time=start_time, end_time=start_time+duration)
    return Animation(position=pos, segments=[segment])
    
def change_animation(pos, newvalue, start_time=0.0, duration=DEFAULT_DURATION):
    segment = AnimationSegment(action=CHANGE, to_value=newvalue, start_time=start_time, end_time=start_time+duration)
    return Animation(position=frompos, segments=[segment])
    
def new_piece_animation(pos, newvalue, start_time=0.0, duration=DEFAULT_DURATION):
    segment = AnimationSegment(action=NEW, from_position=pos, from_value=newvalue, start_time=start_time, end_time=start_time+duration)
    return Animation(position=None, segments=[segment])
    
    
def move_with_capture_animations(frompos, topos, start_time=0.0, duration=DEFAULT_DURATION):
    moveanim = move_animation(frompos, topos, start_time, duration)
    capanim = remove_animation(topos, start_time, duration)
    return [moveanim, capanim]


def test():
    a = AnimationGroup([move_animation( (0,0), (2,4), 0, 1)])
    assert a.duration()==1.0
    
    state = a.animation_state_at_time( (0,0), 1, 0 )
    assert state[0]==(0,0)
    assert state[1][0]==state[1][1]==1
    assert state[2]==1.0
    
    state = a.animation_state_at_time( (3,4), 1, 0.5 )
    assert state[0]==(3,4)
    assert state[1][0]==state[1][1]==1
    assert state[2]==1.0
    
    state = a.animation_state_at_time( (0,0), 1, 0.5 )
    print state
    assert state[0]==(1,2)
    assert state[1][0]==state[1][1]==1
    assert state[2]==1.0
    
    print 'Tests passed'
    
if __name__=='__main__':
    test()
    